from corgidrp.mocks import create_psfsub_dataset,create_ct_cal, gaussian_array
from corgidrp.klip_fm import meas_klip_thrupt, get_closest_psf, inject_psf, measure_noise
from corgidrp.l3_to_l4 import do_psf_subtraction, crop
from corgidrp.astrom import centroid, seppa2xy, create_circular_mask
from corgidrp.data import Image, Dataset
from astropy.io import fits
from scipy.ndimage import shift, rotate
from pyklip.fakes import gaussfit2d


import pytest
import numpy as np
import os

## Helper functions/quantities

lam = 573.8e-9 #m
d = 2.36 #m  
pixscale_mas = 0.0218 * 1000 # mas
fwhm_mas = 1.22 * lam / d * 206265. * 1000.
fwhm_pix = fwhm_mas / pixscale_mas  
res_elem = 3 * fwhm_pix # pix

owa_mas = 450. 
owa_pix = owa_mas / pixscale_mas   

iwa_mas = 140. 
iwa_pix = iwa_mas / pixscale_mas  

# Overriding separations for testing purposes
seps = np.array([10.,15.,20.,25.,30.])

# Mock CT calibration settings

nx,ny = (21,21)
cenx, ceny = (25.,30.)

# Mock PSF subtraction data settings
st_amp = 100.
noise_amp = 1e-3
pl_contrast = 0.0
rolls = [0,10.,0,0]

# Injection test settings
inj_flux = 10.

# KLIP throughput calculation settings
inject_snr = 10.
outdir = 'klipcal_output'
fileprefix = 'FAKE'
annuli = 1
subsections = 1
movement = 1
numbasis = [1]
calibrate_flux = False

# max_thrupt_tolerance = 1 + (noise_amp * (2*fwhm_pix)**2)
max_thrupt_tolerance = 1.05

if not os.path.exists(outdir):
    os.mkdir(outdir)
   
## pyKLIP data class tests

def test_create_ct_cal():
    """Test that mocks.create_ct_cal() generates the correct number of PSFs, 
    each with the correct shape, and that each PSF is a predictable amplitude 
    for debugging purposes.
    """

    nx,ny = (3,5)
    
    cenx,ceny = (10.5,20.5)

    n_psfs = nx * ny

    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    # Check that the correct number of psfs are created
    assert ctcal.ct_excam.shape == (3,n_psfs)

    # Check that each psf is the correct amplitude 
    # (scales with i for debugging purposes)
    for i in range(1,n_psfs+1):
        psf = ctcal.data[i-1]
        assert np.max(psf) == pytest.approx(ctcal.ct_excam[2][i-1],rel=0.01)

        # Check that psf is odd shape and is centered
        assert np.all(np.array(psf.shape) % 2 == 1)
        assert np.array(centroid(psf)) == pytest.approx(np.array(psf.shape)/2.-0.5)


def test_get_closest_psf():
    """Test that the correct PSF is grabbed from the CT Calibration 
    object for each position.
    """
    # Should get the first PSF
    nx,ny = (3,5)
    ct_cenx,ct_ceny = (10.5,20.5)
    cenx,ceny = (10.0,20.0)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=ct_cenx,ceny=ct_ceny,
                  nx=nx,ny=ny)
    
    goal_dxdy = (-100,-100)
    psf = get_closest_psf(ctcal,cenx,ceny,*goal_dxdy)
    assert psf == pytest.approx(ctcal.data[0])

    # Should get the last PSF
    goal_dxdy = (100,100)
    psf = get_closest_psf(ctcal,cenx,ceny,*goal_dxdy)
    assert psf == pytest.approx(ctcal.data[-1])

    # Should get the center PSF (at index 7)
    goal_dxdy = (0.,0.)
    psf = get_closest_psf(ctcal,cenx,ceny,*goal_dxdy)
    assert psf == pytest.approx(ctcal.data[7])

    # Should get the PSF at (9.5,21.5) (at index 3)
    goal_dxdy = (-1.2,1.1)
    psf = get_closest_psf(ctcal,cenx,ceny,*goal_dxdy)
    assert psf == pytest.approx(ctcal.data[3])


def test_inject_psf():
    """Test that PSFs are injected with the correct amplitude and peak pixel location
    """
    # Mock CT cal object
    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny,
                  psfsize=5)

    # Test 0 separation to make sure we can scale and center
    sep = 0.0
    pa = 0.0
    roll = 0.0
    frame_shape_yx = (50,60)
    expected_peak = (ceny,cenx)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)
    
    assert np.max(frame_out.data) == pytest.approx(inj_flux)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak
    
    # Test separation of exactly 1 pixel
    sep = 1.0
    pa = 0.0
    roll = 0.0
    frame_shape_yx = (50,60)
    expected_peak = (ceny+1,cenx)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)
    
    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak

    # Test PA
    sep = 1.0
    pa = 90.0
    roll = 0.0
    frame_shape_yx = (50,60)
    expected_peak = (ceny,cenx-1)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)
    
    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak

    # Test Roll
    sep = 5.0
    pa = 0.0
    roll = 90.0
    frame_shape_yx = (50,60)
    expected_peak = (ceny,cenx+5)

    pri_hdr = fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)
    
    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak

    # Test injecting a psf over the left edge of the array
    sep = 9.0
    pa = 90.0
    roll = 0.0
    frame_shape_yx = (11,21)
    ceny,cenx = (5,10)
    expected_peak_yx = (5,1)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx

    # Test injecting a psf over the right edge of the array
    sep = 9.0
    pa = -90.0
    roll = 0.0
    frame_shape_yx = (11,21)
    ceny,cenx = (5,10)
    expected_peak_yx = (5,19)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx

    
    # Test injecting a psf over the top edge of the array
    sep = 4.0
    pa = 0.0
    roll = 0.0
    frame_shape_yx = (11,21)
    ceny,cenx = (5,10)
    expected_peak_yx = (9,10)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx


    # Test injecting a psf over the bottom edge of the array
    sep = 4.0
    pa = 180.0
    roll = 0.0
    frame_shape_yx = (11,21)
    ceny,cenx = (5,10)
    expected_peak_yx = (1,10)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx


    # Test injecting a psf over the bottom left corner of the array
    sep = 4.0 * np.sqrt(2)
    pa = 135.0
    roll = 0.0
    frame_shape_yx = (11,11)
    ceny,cenx = (5,5)
    expected_peak_yx = (1,1)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx


    # Test injecting a psf over the bottom right corner of the array
    sep = 4.0 * np.sqrt(2)
    pa = -135.0
    roll = 0.0
    frame_shape_yx = (11,11)
    ceny,cenx = (5,5)
    expected_peak_yx = (1,9)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx

    # Test injecting a psf over the top left corner of the array
    sep = 4.0 * np.sqrt(2)
    pa = 45.0
    roll = 0.0
    frame_shape_yx = (11,11)
    ceny,cenx = (5,5)
    expected_peak_yx = (9,1)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx
    
    # Test injecting a psf over the top right corner of the array
    sep = 4.0 * np.sqrt(2)
    pa = -45.0
    roll = 0.0
    frame_shape_yx = (11,11)
    ceny,cenx = (5,5)
    expected_peak_yx = (9,9)

    pri_hdr=fits.Header()
    ext_hdr = fits.Header()
    pri_hdr['ROLL'] = roll
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    frame = Image(np.zeros(frame_shape_yx),
                  pri_hdr=pri_hdr,
                  ext_hdr=ext_hdr)

    frame_out, psf_model, psf_cenxy = inject_psf(frame, ctcal,inj_flux, sep,pa)

    assert np.unravel_index(np.argmax(frame_out.data),frame_out.data.shape) == expected_peak_yx


def test_measure_noise():
    """Check that annular noise profile measurement produces arrays with 
    the correct shape and values for a case with zero noise, and for a case 
    with different noise levels within and without a given radius of pixels.
    Also checks that klip_fm.measure_noise() can return results for only a 
    specific KL mode truncation if desired.
    """

    cenx,ceny = (120.,130.)
    frame_shape_yx = (200,200)
    seps = np.arange(50.,71.,5.)
    border_sep = 62.5
    
    fwhm = 2. # pix

    ext_hdr = fits.Header()
    ext_hdr['STARLOCX'] = cenx
    ext_hdr['STARLOCY'] = ceny
    
    image = np.zeros((2,*frame_shape_yx))

    frame = Image(image,
                  pri_hdr=fits.Header(),
                  ext_hdr=ext_hdr)

    # No KL mode specified
    noise_profile = measure_noise(frame, seps, fwhm)

    # Check data shape
    assert noise_profile.shape == (5,2)
    assert noise_profile == pytest.approx(0.)

    # Different noise values for each KL mode 
    # and inner/outer region

    std1 = 1.
    std2 = 2.
    std3 = 3.
    std4 = 4.

    image = np.zeros((2,*frame_shape_yx))
    frame = Image(image,
                  pri_hdr=fits.Header(),
                  ext_hdr=ext_hdr)

    y, x = np.indices(frame.data.shape[1:])
    sep_map = np.sqrt((y-ceny)**2 + (x-cenx)**2)
    

    rng = np.random.default_rng(0)
    frame.data[0,:,:] = np.where(sep_map>border_sep,rng.normal(0.,std1,frame_shape_yx),rng.normal(0.,std2,frame_shape_yx))
    frame.data[1,:,:] = np.where(sep_map>border_sep,rng.normal(0.,std3,frame_shape_yx),rng.normal(0.,std4,frame_shape_yx))

    # No KL mode specified
    noise_profile = measure_noise(frame, seps, fwhm)

    # Check data shape
    assert noise_profile.shape == (5,2)

    # Check that std measurement is correct
    assert noise_profile[:3,0] == pytest.approx(std2,rel=0.05)
    assert noise_profile[3:,0] == pytest.approx(std1,rel=0.05)
    assert noise_profile[:3,1] == pytest.approx(std4,rel=0.05)
    assert noise_profile[3:,1] == pytest.approx(std3,rel=0.05)
    
    # Specify a particular KL mode
    noise_profile = measure_noise(frame, seps, fwhm,klmode_index=0)

    # Check data shape
    assert noise_profile.shape == (5,)
    assert noise_profile[:3] == pytest.approx(std2,rel=0.05)
    assert noise_profile[3:] == pytest.approx(std1,rel=0.05)

    # Specify a different particular KL mode
    noise_profile = measure_noise(frame, seps, fwhm,klmode_index=1)

    # Check data shape
    assert noise_profile.shape == (5,)
    assert noise_profile[:3] == pytest.approx(std4,rel=0.05)
    assert noise_profile[3:] == pytest.approx(std3,rel=0.05)


def test_meas_klip_ADI():
    """Checks that KLIP throughput measurement for ADI is always between 0 and 1 
    (within 5%) and that throughput increases with separation from the mask.
    """
    global kt_adi 

    mode = 'ADI'
    nsci, nref = (2,0)

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)

    klip_params = {
                'outdir':outdir,'fileprefix':fileprefix,
                'annuli':annuli, 'subsections':subsections, 
                'movement':movement, 'numbasis':numbasis,
                'mode':mode,'calibrate_flux':calibrate_flux}
    
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast)

    # # Plot input dataset
    # import matplotlib.pyplot as plt
    # dataset_frames = mock_sci if mock_ref is None else (*mock_sci,*mock_ref)
    # fig,axes = plt.subplots(1,len(dataset_frames),sharey=True,layout='constrained',figsize=(3*len(dataset_frames),3))
    # for f,frame in enumerate(dataset_frames):
    #     im0 = axes[f].imshow(frame.data,origin='lower')
    #     plt.colorbar(im0,ax=axes[f],shrink=0.8)
    #     axes[f].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'],s=1)
    #     name = 'Sci' if f < len(mock_sci) else 'Ref'
    #     n_frame = f if f < len(mock_sci) else f-len(mock_sci)
    #     axes[f].set_title(f'{name} Input Frame {n_frame}')
    # plt.show()

    psfsub_dataset = do_psf_subtraction(mock_sci,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=mode,
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)
    
    # # Plot Psf subtraction result
    # if psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'RDI':
    #     analytical_result = rotate(mock_sci[0].data - mock_ref[0].data,-rolls[0],reshape=False,cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI':
    #     analytical_result = shift((rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2,
    #                     [0.5,0.5],
    #                     cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI+RDI':
    #     analytical_result = (rotate(mock_sci[0].data - (mock_sci[1].data/2+mock_ref[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - (mock_sci[0].data/2+mock_ref[0].data/2),-rolls[1],reshape=False,cval=0)) / 2
    # import matplotlib.pyplot as plt
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(psfsub_dataset[0].data[0],origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Output data')
    # im1 = axes[1].imshow(analytical_result,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Analytical result')
    # diff = psfsub_dataset[0].data[0] - analytical_result
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Difference')
    # plt.suptitle(f'PSF Subtraction {psfsub_dataset[0].pri_hdr["KLIP_ALG"]} ({psfsub_dataset[0].ext_hdr["KLMODE0"]} KL Modes)')
    # plt.show()   

    klip_params['mode'] = mode
    kt_adi = meas_klip_thrupt(mock_sci, mock_ref, # pre-psf-subtracted dataset
                     psfsub_dataset, # post-subtraction dataset
                     ctcal,
                     klip_params,
                     inject_snr,
                     seps = seps, # in pixels from mask center
                     cand_locs=[])

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots(figsize=(6,4))
    # plt.plot(kt_adi[0],kt_adi[1],label='ADI')
    # plt.title('KLIP throughput')
    # plt.legend()
    # plt.xlabel('separation (pixels)')
    # plt.show()

    # Check KL thrupt is <= 1 within noise tolerance
    assert np.all(kt_adi[1:,:,0] < max_thrupt_tolerance)

    # Check KL thrupt is > 0
    assert np.all(kt_adi[1:,:,0] > 0.)

    # Check KL thrupt increases with separation
    for i in range(1,len(kt_adi[0])):
        assert np.all(kt_adi[1:,i,0] > kt_adi[1:,i-1,0] - 0.05) # add a fudge factor of 5%

    # Check recovered FWHM increases with separation
    for i in range(1,len(kt_adi[0])):
        assert np.all(kt_adi[1:,i,1] > kt_adi[1:,i-1,1] - 0.1) # add fudge factor of 0.1 pixels


def test_meas_klip_RDI():
    """Checks that KLIP throughput measurement for RDI is always between 0.8 and 1 
    (within 5%).
    """
    global kt_rdi
    mode = 'RDI'
    nsci, nref = (1,1)

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)

    klip_params = {
                'outdir':outdir,'fileprefix':fileprefix,
                'annuli':annuli, 'subsections':subsections, 
                'movement':movement, 'numbasis':numbasis,
                'mode':mode,'calibrate_flux':calibrate_flux}
    
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast)

    # # Plot input dataset
    # import matplotlib.pyplot as plt
    # dataset_frames = mock_sci if mock_ref is None else (*mock_sci,*mock_ref)
    # fig,axes = plt.subplots(1,len(dataset_frames),sharey=True,layout='constrained',figsize=(3*len(dataset_frames),3))
    # for f,frame in enumerate(dataset_frames):
    #     im0 = axes[f].imshow(frame.data,origin='lower')
    #     plt.colorbar(im0,ax=axes[f],shrink=0.8)
    #     axes[f].scatter(frame.ext_hdr['STARLOCX'],frame.ext_hdr['STARLOCY'],s=1)
    #     name = 'Sci' if f < len(mock_sci) else 'Ref'
    #     n_frame = f if f < len(mock_sci) else f-len(mock_sci)
    #     axes[f].set_title(f'{name} Input Frame {n_frame}')
    # plt.show()

    psfsub_dataset = do_psf_subtraction(mock_sci,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=mode,
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)

    klip_params['mode'] = mode
    kt_rdi = meas_klip_thrupt(mock_sci, mock_ref, # pre-psf-subtracted dataset
                     psfsub_dataset, # post-subtraction dataset
                     ctcal,
                     klip_params,
                     inject_snr,
                     seps=seps)

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots(figsize=(6,4))
    # plt.plot(kt_rdi[0],kt_rdi[1],label='RDI')
    # plt.title('KLIP throughput')
    # plt.legend()
    # plt.xlabel('separation (pixels)')
    # plt.show()

    # Check KL thrupt is <= 1 within noise tolerance
    assert np.all(kt_rdi[1:,:,0] < max_thrupt_tolerance)

    # Check KL thrupt > 0.8
    assert np.all(kt_rdi[1:,:,0] > 0.8)


def test_meas_klip_ADIRDI():
    """Checks that KLIP throughput measurement for ADI+RDI is always between 0 and 1.
    """
    global kt_adirdi
    
    mode = 'ADI+RDI'
    nsci, nref = (2,1)

    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    klip_params = {
                'outdir':outdir,'fileprefix':fileprefix,
                'annuli':annuli, 'subsections':subsections, 
                'movement':movement, 'numbasis':numbasis,
                'mode':mode,'calibrate_flux':calibrate_flux}
    
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast)

    psfsub_dataset = do_psf_subtraction(mock_sci,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=mode,
                                do_crop=False,
                                measure_klip_thrupt=False,
                                measure_1d_core_thrupt=False)

    kt_adirdi = meas_klip_thrupt(mock_sci, mock_ref, # pre-psf-subtracted dataset
                     psfsub_dataset, # post-subtraction dataset
                     ctcal,
                     klip_params,
                     inject_snr,
                     seps=seps)

    # Check KL thrupt is <= 1 within noise tolerance
    assert np.all(kt_adirdi[1:,:,0] < max_thrupt_tolerance)

    # Check KL thrupt is > 0
    assert np.all(kt_adirdi[1:,:,0] > 0.)


def test_compare_RDI_ADI():
    """Check that mean ADI throughput is lower than mean RDI throughput.
    """

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots()
    # ax.plot(kt_adi[0],kt_adi[1],label='ADI')
    # ax.plot(kt_rdi[0],kt_adirdi[1],label='ADI+RDI')
    # ax.plot(kt_rdi[0],kt_rdi[1],label='RDI')
    # plt.legend()
    # ax.set_ylim(-0.1,1.1)
    # plt.title('KLIP Throughput')
    # plt.xlabel('Separation (pixels)')
    # plt.show()

    # Check that ADI thrupt < RDI thrupt
    mean_adi = np.mean(kt_adi[1:,:,0])
    mean_rdi = np.mean(kt_rdi[1:,:,0])

    assert mean_adi < mean_rdi


def test_psfsub_withklipandctmeas_adi():
    """Check that KLIP throughput and CT calibration can be run as part 
    of the PSF subtraction step function. Check that KLIP throughput and 
    core throughput sample the same separations by default. Check that an 
    input planet flux (after CT) can be recovered via the KLIP throughput 
    within 1% error after ADI.
    """
    nsci, nref = (2,0) # Mode ADI
 
    st_amp = 100.
    noise_amp = 1e-3
    pl_contrast = 3e-4 
    pl_amp = st_amp * pl_contrast
    pl_loc = (20.,0.)
    est_pl_snr = pl_amp / noise_amp
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast,
                                            pl_sep=pl_loc[0],
                                )

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    
    psfsub_dataset = do_psf_subtraction(mock_sci,ctcal,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=None,
                                do_crop=False,
                                measure_klip_thrupt=True,
                                measure_1d_core_thrupt=True,
                                cand_locs= [pl_loc],
                                kt_seps=[pl_loc[0]])
    

    # Plot Psf subtraction result
    # if psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'RDI':
    #     analytical_result = rotate(mock_sci[0].data - mock_ref[0].data,-rolls[0],reshape=False,cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI':
    #     analytical_result = shift((rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2,
    #                     [0.5,0.5],
    #                     cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI+RDI':
    #     analytical_result = (rotate(mock_sci[0].data - (mock_sci[1].data/2+mock_ref[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - (mock_sci[0].data/2+mock_ref[0].data/2),-rolls[1],reshape=False,cval=0)) / 2
    # import matplotlib.pyplot as plt
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(psfsub_dataset[0].data[0],origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Output data')
    # im1 = axes[1].imshow(analytical_result,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Analytical result')
    # diff = psfsub_dataset[0].data[0] - analytical_result
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Difference')
    # plt.suptitle(f'PSF Subtraction {psfsub_dataset[0].pri_hdr["KLIP_ALG"]} ({psfsub_dataset[0].ext_hdr["KLMODE0"]} KL Modes)')
    # plt.show()    

    # Check that klip and ct separations are the same
    kt = psfsub_dataset[0].hdu_list['KL_THRU'].data
    kt_seps = kt[0,:,0]

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots(figsize=(6,4))
    # plt.scatter(kt[0],kt[1],label=psfsub_dataset[0].pri_hdr["KLIP_ALG"])
    # plt.title('KLIP throughput')
    # plt.legend()
    # plt.xlabel('separation (pixels)')
    # plt.show()

    ct = psfsub_dataset[0].hdu_list['CT_THRU'].data
    ct_seps = ct[0]

    assert np.all(kt_seps == ct_seps)

    # Fit 2d gaussian to planet location:

    # Crop data around location to be same as psf_model cutout
    locxy = seppa2xy(*pl_loc,psfsub_dataset[0].ext_hdr['STARLOCX'],psfsub_dataset[0].ext_hdr['STARLOCY'])

    # Measure background via sigma clip 
    n_loops = 5
    masked_data = psfsub_dataset[0].data.copy()[0]
    for n in range(n_loops):
        std = np.nanstd(masked_data)
        med = np.nanmedian(masked_data)
        clip_thresh = 3 * std
        masked_data = np.where(np.abs(masked_data-med)>clip_thresh,np.nan,masked_data)
    
    # Subtract median
    bg_level = np.nanmedian(masked_data)
    medsubtracted_data = psfsub_dataset[0].data[0] - bg_level

    # Crop the data, pad with nans if we're cropping over the edge
    cutout_shape = np.array([21,21])
    cutout = np.zeros(cutout_shape)
    cutoutcenyx = cutout_shape/2. - 0.5
    cutout[:] = np.nan
    cutout_starty, cutout_startx = (0,0)
    cutout_endy, cutout_endx = cutout.shape

    data_shape = medsubtracted_data.shape
    data_center_indyx = np.array([locxy[1],locxy[0]]).astype(int)
    data_start_indyx = (data_center_indyx - cutout_shape//2)
    data_end_indyx = (data_start_indyx + cutout_shape)
    data_starty,data_startx = data_start_indyx
    data_endy,data_endx = data_end_indyx
    
    if data_starty < 0:
        cutout_starty = -data_starty
        data_starty = 0
    
    if data_startx < 0:
        cutout_startx = -data_startx
        data_startx = 0
    
    if data_endy >= data_shape[0]:
        y_overhang = data_endy - medsubtracted_data.shape[0]
        cutout_endy = cutout_shape[0] - y_overhang
        data_endy = data_shape[0]

    if data_endx >= data_shape[1]:
        x_overhang = data_endx - medsubtracted_data.shape[1]
        cutout_endx = cutout_shape[1] - x_overhang
        data_endx = data_shape[1]

    cutout[cutout_starty:cutout_endy,
                cutout_startx:cutout_endx] = medsubtracted_data[data_starty:data_endy,
                                                    data_startx:data_endx]
    
    postklip_peak, post_fwhm, post_xfit, post_yfit = gaussfit2d(
                cutout, 
                cutoutcenyx[1], 
                cutoutcenyx[0], 
                searchrad=5, 
                guessfwhm=fwhm_pix, 
                guesspeak=pl_amp, 
                refinefit=True) 

    # import matplotlib.pyplot as plt
    # post_sigma = post_fwhm / (2 * np.sqrt(2. * np.log(2.)))
    # final_model = gaussian_array(array_shape=cutout_shape,
    #                              sigma=post_sigma,
    #                              amp=postklip_peak,
    #                              xoffset=post_xfit-10.,
    #                              yoffset=post_yfit-10.)
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(cutout,origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Data')
    # im1 = axes[1].imshow(final_model,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Model')
    # diff = cutout-final_model
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Residuals')
    # plt.suptitle(f'Final PSF Fit')
    # plt.show()    

    
    pl_kt = kt[1,np.argmin(np.abs(kt_seps-pl_loc[0])),0]
    
    pl_counts = np.pi * pl_amp * fwhm_pix**2 / 4. / np.log(2.)
    recovered_pl_counts = np.pi * postklip_peak * post_fwhm**2 / 4. / np.log(2.)
    recovered_pl_counts_ktcorrected = recovered_pl_counts / pl_kt


    assert pl_counts == pytest.approx(recovered_pl_counts_ktcorrected,rel = 0.01) 


def test_psfsub_withklipandctmeas_rdi():
    """Check that KLIP throughput and CT calibration can be run as part 
    of the PSF subtraction step function. Check that KLIP throughput and 
    core throughput sample the same separations by default. Check that an 
    input planet flux (after CT) can be recovered via the KLIP throughput 
    within 5% error after RDI.
    """
    nsci, nref = (1,1) # Mode RDI
 
    st_amp = 100.
    noise_amp = 1e-3
    pl_contrast = 3e-3 
    pl_amp = st_amp * pl_contrast
    pl_loc = (20.,0.)
    est_pl_snr = pl_amp / noise_amp
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast,
                                            pl_sep=pl_loc[0],
                                )

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    
    psfsub_dataset = do_psf_subtraction(mock_sci,ctcal,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=None,
                                do_crop=False,
                                measure_klip_thrupt=True,
                                measure_1d_core_thrupt=True,
                                cand_locs= [pl_loc],
                                kt_seps=[pl_loc[0]])
    

    # Plot Psf subtraction result
    
    if psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'RDI':
        analytical_result = rotate(mock_sci[0].data - mock_ref[0].data,-rolls[0],reshape=False,cval=np.nan)
    elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI':
        analytical_result = shift((rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2,
                        [0.5,0.5],
                        cval=np.nan)
    elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI+RDI':
        analytical_result = (rotate(mock_sci[0].data - (mock_sci[1].data/2+mock_ref[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - (mock_sci[0].data/2+mock_ref[0].data/2),-rolls[1],reshape=False,cval=0)) / 2

    mask = create_circular_mask(analytical_result.shape[-2:],
                                r=3*fwhm_pix,
                                center=(psfsub_dataset[0].ext_hdr['STARLOCX'],
                                        psfsub_dataset[0].ext_hdr['STARLOCY']))
    masked_analytical_result = np.where(mask,np.nan,analytical_result)


    # Check that klip and ct separations are the same
    kt = psfsub_dataset[0].hdu_list['KL_THRU'].data
    kt_seps = kt[0,:,0]

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots(figsize=(6,4))
    # plt.scatter(kt[0],kt[1],label=psfsub_dataset[0].pri_hdr["KLIP_ALG"])
    # plt.title('KLIP throughput')
    # plt.legend()
    # plt.xlabel('separation (pixels)')
    # plt.show()

    ct = psfsub_dataset[0].hdu_list['CT_THRU'].data
    ct_seps = ct[0]

    assert np.all(kt_seps == ct_seps)

    # Fit 2d gaussian to planet location:

    # Crop data around location to be same as psf_model cutout
    locxy = seppa2xy(*pl_loc,psfsub_dataset[0].ext_hdr['STARLOCX'],psfsub_dataset[0].ext_hdr['STARLOCY'])

    # Measure background via sigma clip 
    n_loops = 5
    masked_data = psfsub_dataset[0].data.copy()[0]
    for n in range(n_loops):
        std = np.nanstd(masked_data)
        med = np.nanmedian(masked_data)
        clip_thresh = 3 * std
        masked_data = np.where(np.abs(masked_data-med)>clip_thresh,np.nan,masked_data)
    
    # Subtract median
    bg_level = np.nanmedian(masked_data)
    medsubtracted_data = psfsub_dataset[0].data[0] #- bg_level

    # import matplotlib.pyplot as plt
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(medsubtracted_data,origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'BG-subtracted Output data')
    # im1 = axes[1].imshow(analytical_result,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Analytical result')
    # diff = medsubtracted_data - masked_analytical_result
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Difference')
    # plt.suptitle(f'PSF Subtraction {psfsub_dataset[0].pri_hdr["KLIP_ALG"]} ({psfsub_dataset[0].ext_hdr["KLMODE0"]} KL Modes)')
    # plt.show()    


    # Crop the data, pad with nans if we're cropping over the edge
    cutout_shape = np.array([21,21])
    cutout = np.zeros(cutout_shape)
    cutoutcenyx = cutout_shape/2. - 0.5
    cutout[:] = np.nan
    cutout_starty, cutout_startx = (0,0)
    cutout_endy, cutout_endx = cutout.shape

    data_shape = medsubtracted_data.shape
    data_center_indyx = np.array([locxy[1],locxy[0]]).astype(int)
    data_start_indyx = (data_center_indyx - cutout_shape//2)
    data_end_indyx = (data_start_indyx + cutout_shape)
    data_starty,data_startx = data_start_indyx
    data_endy,data_endx = data_end_indyx
    
    if data_starty < 0:
        cutout_starty = -data_starty
        data_starty = 0
    
    if data_startx < 0:
        cutout_startx = -data_startx
        data_startx = 0
    
    if data_endy >= data_shape[0]:
        y_overhang = data_endy - medsubtracted_data.shape[0]
        cutout_endy = cutout_shape[0] - y_overhang
        data_endy = data_shape[0]

    if data_endx >= data_shape[1]:
        x_overhang = data_endx - medsubtracted_data.shape[1]
        cutout_endx = cutout_shape[1] - x_overhang
        data_endx = data_shape[1]

    cutout[cutout_starty:cutout_endy,
                cutout_startx:cutout_endx] = medsubtracted_data[data_starty:data_endy,
                                                    data_startx:data_endx]
    

    postklip_peak, post_fwhm, post_xfit, post_yfit = gaussfit2d(
                cutout, 
                cutoutcenyx[1], 
                cutoutcenyx[0], 
                searchrad=5, 
                guessfwhm=fwhm_pix, 
                guesspeak=pl_amp, 
                refinefit=True) 
    post_sigma = post_fwhm / (2 * np.sqrt(2. * np.log(2.)))
    final_model = gaussian_array(array_shape=cutout_shape,
                                 sigma=post_sigma,
                                 amp=postklip_peak,
                                 xoffset=post_xfit-10.,
                                 yoffset=post_yfit-10.)
    

    diff = cutout-final_model

    # import matplotlib.pyplot as plt
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(cutout,origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Data')
    # im1 = axes[1].imshow(final_model,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Model')
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Residuals')
    # plt.suptitle(f'Final PSF Fit')
    # plt.show()    
    
    pl_kt = kt[1,np.argmin(np.abs(kt_seps-pl_loc[0])),0]
    
    pl_counts = np.pi * pl_amp * fwhm_pix**2 / 4. / np.log(2.)
    recovered_pl_counts = np.pi * postklip_peak * post_fwhm**2 / 4. / np.log(2.)
    recovered_pl_counts_ktcorrected = recovered_pl_counts / pl_kt

    assert pl_counts == pytest.approx(recovered_pl_counts_ktcorrected,rel = 0.05) 


def test_psfsub_withKTandCTandCrop_adi():
    """Check that KLIP throughput, CT calibration, and crop step can be run as part 
    of the PSF subtraction step function. Check that KLIP throughput and 
    core throughput sample the same separations by default. Check that an 
    input planet flux (after CT) can be recovered via the KLIP throughput 
    within 1% error after ADI.
    """    
    nsci, nref = (2,0) # Mode ADI
 
    st_amp = 100.
    noise_amp = 1e-3
    pl_contrast = 3e-4 
    pl_amp = st_amp * pl_contrast
    pl_loc = (20.,0.)
    mock_sci,mock_ref = create_psfsub_dataset(nsci,nref,rolls,
                                            fwhm_pix=fwhm_pix,
                                            st_amp=st_amp,
                                            noise_amp=noise_amp,
                                            pl_contrast=pl_contrast,
                                            pl_sep=pl_loc[0],
                                )

    nx,ny = (21,21)
    cenx, ceny = (25.,30.)
    ctcal = create_ct_cal(fwhm_mas, cfam_name='1F',
                  cenx=cenx,ceny=ceny,
                  nx=nx,ny=ny)
    
    
    psfsub_dataset = do_psf_subtraction(mock_sci,ctcal,
                                reference_star_dataset=mock_ref,
                                numbasis=numbasis,
                                fileprefix='test_KL_THRU',
                                mode=None,
                                do_crop=True,
                                measure_klip_thrupt=True,
                                measure_1d_core_thrupt=True,
                                cand_locs= [pl_loc],
                                kt_seps=[pl_loc[0]])

    # # Plot Psf subtraction result
    # import matplotlib.pyplot as plt
    # if psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'RDI':
    #     analytical_result = rotate(mock_sci[0].data - mock_ref[0].data,-rolls[0],reshape=False,cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI':
    #     analytical_result = shift((rotate(mock_sci[0].data - mock_sci[1].data,-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - mock_sci[0].data,-rolls[1],reshape=False,cval=0)) / 2,
    #                     [0.5,0.5],
    #                     cval=np.nan)
    # elif psfsub_dataset[0].pri_hdr['KLIP_ALG'] == 'ADI+RDI':
    #     analytical_result = (rotate(mock_sci[0].data - (mock_sci[1].data/2+mock_ref[0].data/2),-rolls[0],reshape=False,cval=0) + rotate(mock_sci[1].data - (mock_sci[0].data/2+mock_ref[0].data/2),-rolls[1],reshape=False,cval=0)) / 2
    # prihdr = fits.Header()
    # exthdr = fits.Header()
    # exthdr["STARLOCX"] = mock_sci[0].ext_hdr["STARLOCX"]
    # exthdr["STARLOCY"] = mock_sci[0].ext_hdr["STARLOCY"]
    # exthdr["STARLOCX"] = mock_sci[0].ext_hdr["STARLOCX"]
    # exthdr["STARLOCY"] = mock_sci[0].ext_hdr["STARLOCY"]
    # exthdr.set("LSAMNAME",'NFOV')
    # ar_image = Image(analytical_result,
    #                  prihdr,
    #                  exthdr) 
    # ar_dataset_cropped = crop(Dataset([ar_image]))
    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # im0 = axes[0].imshow(psfsub_dataset[0].data[0],origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Output data')
    # im1 = axes[1].imshow(ar_dataset_cropped[0].data,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Analytical result')
    # diff = psfsub_dataset[0].data[0] - ar_dataset_cropped[0].data
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Difference')
    # plt.suptitle(f'PSF Subtraction {psfsub_dataset[0].pri_hdr["KLIP_ALG"]} ({psfsub_dataset[0].ext_hdr["KLMODE0"]} KL Modes)')
    # plt.show()    

    # Check that klip and ct separations are the same
    kt = psfsub_dataset[0].hdu_list['KL_THRU'].data
    kt_seps = kt[0,:,0]

    # import matplotlib.pyplot as plt
    # fig,ax = plt.subplots(figsize=(6,4))
    # plt.scatter(kt[0],kt[1],label=psfsub_dataset[0].pri_hdr["KLIP_ALG"])
    # plt.title('KLIP throughput')
    # plt.legend()
    # plt.xlabel('separation (pixels)')
    # plt.show()

    ct = psfsub_dataset[0].hdu_list['CT_THRU'].data
    ct_seps = ct[0]

    assert np.all(kt_seps == ct_seps)

    # Fit 2d gaussian to planet location:

    # Crop data around location to be same as psf_model cutout
    locxy = seppa2xy(*pl_loc,psfsub_dataset[0].ext_hdr['STARLOCX'],psfsub_dataset[0].ext_hdr['STARLOCY'])

    # Measure background via sigma clip 
    n_loops = 5
    masked_data = psfsub_dataset[0].data.copy()[0]
    for n in range(n_loops):
        std = np.nanstd(masked_data)
        med = np.nanmedian(masked_data)
        clip_thresh = 3 * std
        masked_data = np.where(np.abs(masked_data-med)>clip_thresh,np.nan,masked_data)
    
    # Subtract median
    bg_level = np.nanmedian(masked_data)
    medsubtracted_data = psfsub_dataset[0].data[0] - bg_level

    # Crop the data, pad with nans if we're cropping over the edge
    cutout_shape = np.array([21,21])
    cutout = np.zeros(cutout_shape)
    cutoutcenyx = cutout_shape/2. - 0.5
    cutout[:] = np.nan
    cutout_starty, cutout_startx = (0,0)
    cutout_endy, cutout_endx = cutout.shape

    data_shape = medsubtracted_data.shape
    data_center_indyx = np.array([locxy[1],locxy[0]]).astype(int)
    data_start_indyx = (data_center_indyx - cutout_shape//2)
    data_end_indyx = (data_start_indyx + cutout_shape)
    data_starty,data_startx = data_start_indyx
    data_endy,data_endx = data_end_indyx
    
    if data_starty < 0:
        cutout_starty = -data_starty
        data_starty = 0
    
    if data_startx < 0:
        cutout_startx = -data_startx
        data_startx = 0
    
    if data_endy >= data_shape[0]:
        y_overhang = data_endy - medsubtracted_data.shape[0]
        cutout_endy = cutout_shape[0] - y_overhang
        data_endy = data_shape[0]

    if data_endx >= data_shape[1]:
        x_overhang = data_endx - medsubtracted_data.shape[1]
        cutout_endx = cutout_shape[1] - x_overhang
        data_endx = data_shape[1]

    cutout[cutout_starty:cutout_endy,
                cutout_startx:cutout_endx] = medsubtracted_data[data_starty:data_endy,
                                                    data_startx:data_endx]
    

    postklip_peak, post_fwhm, post_xfit, post_yfit = gaussfit2d(
                cutout, 
                cutoutcenyx[1], 
                cutoutcenyx[0], 
                searchrad=5, 
                guessfwhm=fwhm_pix, 
                guesspeak=pl_amp, 
                refinefit=True) 

    # fig,axes = plt.subplots(1,3,sharey=True,layout='constrained',figsize=(12,3))
    # post_sigma = post_fwhm / (2 * np.sqrt(2. * np.log(2.)))
    # final_model = gaussian_array(array_shape=cutout_shape,
    #                              sigma=post_sigma,
    #                              amp=postklip_peak,
    #                              xoffset=post_xfit-10.,
    #                              yoffset=post_yfit-10.)
    # im0 = axes[0].imshow(cutout,origin='lower')
    # plt.colorbar(im0,ax=axes[0],shrink=0.8)
    # axes[0].set_title(f'Data')
    # im1 = axes[1].imshow(final_model,origin='lower')
    # plt.colorbar(im1,ax=axes[1],shrink=0.8)
    # axes[1].set_title('Model')
    # diff = cutout-final_model
    # im2 = axes[2].imshow(diff,origin='lower')
    # plt.colorbar(im2,ax=axes[2],shrink=0.8)
    # axes[2].set_title('Residuals')
    # plt.suptitle(f'Final PSF Fit')
    # plt.show()    

    pl_kt = kt[1,np.argmin(np.abs(kt_seps-pl_loc[0])),0]
    
    pl_counts = np.pi * pl_amp * fwhm_pix**2 / 4. / np.log(2.)
    recovered_pl_counts = np.pi * postklip_peak * post_fwhm**2 / 4. / np.log(2.)
    recovered_pl_counts_ktcorrected = recovered_pl_counts / pl_kt

    assert pl_counts == pytest.approx(recovered_pl_counts_ktcorrected,rel=0.01)


if __name__ == '__main__':  
    test_create_ct_cal()
    test_get_closest_psf()
    test_inject_psf()
    test_measure_noise()

    test_meas_klip_ADI()
    test_meas_klip_RDI()
    test_meas_klip_ADIRDI()
    test_compare_RDI_ADI()

    # test_psfsub_withklipandctmeas_adi()
    # test_psfsub_withklipandctmeas_rdi()
    # test_psfsub_withKTandCTandCrop_adi()
    



#A file to test the kgain conversion
import os
import pickle
import numpy as np
from corgidrp.mocks import create_default_L1_headers
import corgidrp.data as data
import corgidrp.l2a_to_l2b as l2a_to_l2b
import pytest
import astropy.io.fits as fits

def test_kgain():
    """
    test the KGain class and the calibration file and the unit conversion
    """
    #test KGain class and cal file
    prhd, exthd = create_default_L1_headers()
    dat = np.ones([1024,1024]) * 2
    err = np.ones([1,1024,1024]) * 0.5
    ptc = np.ones([2,1024])
    exthd["RN"] = 100
    exthd["RN_ERR"] = 2
    ptc_hdr = fits.Header()
    image1 = data.Image(dat,pri_hdr = prhd, ext_hdr = exthd, err = err)
    image2 = image1.copy()
    image1.filename = "test1_L1_.fits"
    image2.filename = "test2_L1_.fits"
    dataset= data.Dataset([image1, image2])

    gain_value = np.array([[9.55]])
    gain_err = np.array([[[1.]]])
    kgain = data.KGain(gain_value, pri_hdr = prhd, ext_hdr = exthd, input_dataset = dataset)
    assert kgain.filename.split(".")[0] == "test2_KRN_CAL"
    assert kgain.value == gain_value[0,0]
    assert kgain.data[0,0] == gain_value[0,0]
    
    #test ptc and error extension
    kgain_ptc = data.KGain(gain_value, err = gain_err, ptc = ptc, pri_hdr = prhd, ext_hdr = exthd, ptc_hdr = ptc_hdr, input_dataset = dataset)
    assert kgain_ptc.error == gain_err[0,0]
    assert kgain_ptc.ptc[0,0] == 1.
    assert kgain_ptc.ptc_hdr is not None
    
    # check the kgain can be pickled (for CTC operations)
    pickled = pickle.dumps(kgain_ptc)
    pickled_kgain = pickle.loads(pickled)
    assert np.all((kgain_ptc.data == pickled_kgain.data))

    #test copy and save
    kgain_ptc_copy = kgain_ptc.copy(copy_data = False)
    assert kgain_ptc_copy.value == gain_value[0,0]
    kgain_ptc_copy = kgain_ptc.copy()
    assert kgain_ptc_copy.value == gain_value[0,0]
    
    assert kgain_ptc_copy.ptc[0,0] == 1.
    assert kgain_ptc_copy.error == gain_err[0,0]
    assert kgain_ptc_copy.ptc_hdr is not None
    assert kgain_ptc_copy.err_hdr is not None
    
    # save KGain
    calibdir = os.path.join(os.path.dirname(__file__), "testcalib")
    kgain_filename = "kgain_calib.fits"
    if not os.path.exists(calibdir):
        os.mkdir(calibdir)
    kgain_ptc.save(filedir=calibdir, filename=kgain_filename)        
        
    kgain_filepath = os.path.join(calibdir, kgain_filename)
    kgain_open = data.KGain(kgain_filepath)
    assert kgain_open.value == gain_value[0,0]
    assert kgain_open.error == gain_err[0,0]
    assert kgain_open.ptc[0,0] == 1.
    assert kgain_open.ptc_hdr["EXTNAME"] == "PTC"
    assert kgain_open.err_hdr is not None
    
    # check the kgain can be pickled (for CTC operations)
    pickled = pickle.dumps(kgain_open)
    pickled_kgain = pickle.loads(pickled)
    assert np.all((kgain_open.data == pickled_kgain.data))

    # test convert_to_electrons
    k_gain = kgain.value

    gain_dataset = l2a_to_l2b.convert_to_electrons(dataset, kgain)

    assert np.mean(gain_dataset[0].data) == pytest.approx(k_gain * np.mean(dataset[0].data), abs = 1e-4)
    assert np.mean(gain_dataset[0].err) == pytest.approx(k_gain * np.mean(dataset[0].err), abs = 1e-4)

    #test header updates
    assert gain_dataset[0].ext_hdr["BUNIT"] == "detected EM electrons"
    assert gain_dataset[0].err_hdr["BUNIT"] == "detected EM electrons"
    assert gain_dataset[0].ext_hdr["KGAINPAR"] == k_gain
    assert gain_dataset[0].ext_hdr["KGAIN_ER"] == kgain.error[0]
    assert gain_dataset[0].ext_hdr["RN"] > 0
    assert gain_dataset[0].ext_hdr["RN_ERR"] > 0
    assert("converted" in str(gain_dataset[0].ext_hdr["HISTORY"]))

if __name__ == "__main__":
    test_kgain()



import os
import copy
import pytest
import random
import numpy as np
from pathlib import Path
import astropy.io.fits as fits

from corgidrp import sorting as sorting
import corgidrp.data as data
from corgidrp.data import Image
from corgidrp.mocks import create_default_L1_headers

# Functions
def get_cmdgain_exptime_mean_frame(
    exptime_sec=None,
    nframes=None,
    ):
    """
    Create an array of CMDGAIN, EXPTIME for frames that will be used to
    generate a mean frame.

    Args:
      exptime_sec (float): exposure time of the frame in seconds
      nframes (int): (minimum) number of frames to generate a mean frame

    Returns:
      cmdgain_list (list): list of commanded gains
      exptime_list (list): list of exposure frames
    """
    if exptime_sec is None:
        raise Exception('Missing input exposure times for mean frame data')
    if nframes is None:
        raise Exception('Missing number of frames for mean frame data')

    cmdgain_list = [1] * nframes
    exptime_list = [exptime_sec] * nframes
    return cmdgain_list, exptime_list

def get_cmdgain_exptime_kgain(
    exptime_sec=None,
    nframes=None,
    ):
    """
    Create an array of CMDGAIN, EXPTIME for frames that will be used to
    calibrate k-gain.

    Args:
      exptime_sec (list): set of distinct exposure times n seconds chosen to
      collect frames for K-gain calibration
      nframes (int): number of frames per ditinct exposure time

    Returns:
      cmdgain_list (list): list of commanded gains
      exptime_list (list): list of exposure frames
    """
    if exptime_sec is None:
        raise Exception('Missing input exposure times for K=gain calibration data')
    if nframes is None:
        raise Exception('Missing input number of frames for K=gain calibration data')

    cmdgain_list = [1] * (len(exptime_sec) * nframes)
    exptime_list = []
    for exptime in exptime_sec:
        exptime_list += [exptime] * nframes
    return cmdgain_list, exptime_list

def get_cmdgain_exptime_nonlin(
    exptime_sec=None,
    nonunity_em=None,
    change_exptime=False,
    ):
    """
    Create an array of CMDGAIN, EXPTIME for frames that will be used to
    calibrate non-linearity.

    Args:
      exptime_sec (list): set of distinct exposure times in seconds chosen to
        collect frames for non-linearity calibration for each EM gain
      nonunity_em (list): set of ditinct (non-unity) EM gains chosen to collect
        data for non-linearity
      change_exptime (bool) (optional): if True, it will change the input exposure
        times by a small amount without changing the ordering of exptime_sec

    Returns:
      cmdgain_list (list): list of commanded gains
      exptime_list (list): list of exposure frames
    """
    if exptime_sec is None:
        raise Exception('Missing input exposure times for non-linearity calibration data')
    if nonunity_em is None:
        raise Exception('Missing input EM gain for non-linearity calibration data')

    cmdgain_list = []
    exptime_list = []
    fac_change = 0
    if change_exptime:
        # Avoid the (unlikely) coincidence of +1/-1 in the uniform distribution
        fac_change = min(np.abs(np.diff(exptime_sec))) / 3
    for emgain in nonunity_em:
        cmdgain_list += [emgain] * len(exptime_sec)
        exptime_sec = (np.array(exptime_sec) *
            (1 + fac_change*random.uniform(-1, 1)))
        exptime_list += exptime_sec.tolist()
    return cmdgain_list, exptime_list

def get_cmdgain_exptime_emgain(
    em_emgain=None,
    exptime_emgain_sec=None,
    nframes=None,
    ):
    """
    Create an array of CMDGAIN, EXPTIME for frames that will be used to
    calibrate EM-gain vs DAC. Notice the pairing between unity and non-unity
    gain frames.

     Args:
      em_emgain (list): set of ditinct (non-unity) EM gains chosen to collect
        data for EM gain with pupil images
      exptime_emgain_sec (list): set of distinct exposure times in seconds chosen to
        collect frames for non-linearity calibration for each EM gain
      nframes (int): number of frames per ditinct exposure time

    Returns:
      cmdgain_list (list): list of commanded gains
      exptime_list (list): list of exposure frames
    """
    if em_emgain is None:
        raise Exception('Missing input EM gain for EM-gain calibration data')
    if exptime_emgain_sec is None:
        raise Exception('Missing input exposure times for EM-gain calibration data')
    if nframes is None:
        raise Exception('Missing input number of frames for EM-gain calibration data')

    # Create pairs of frames
    cmdgain_list = []
    for idx in range(len(em_emgain)):
        cmdgain_list += [em_emgain[idx]] * nframes[idx]
    exptime_list = []
    for idx in range(len(exptime_emgain_sec)):
        exptime_list += [exptime_emgain_sec[idx]] * nframes[idx]
    # Unity and non-unity gains
    return cmdgain_list, exptime_list

def make_minimal_image(
    cmdgain=1,
    exptime_sec=0,
    frameid=0,
        ):
    """
    This function makes a mock frame with minimum memory in its data and error
    fields. It is used in this test script only.

    Args:
      cmdgain (float): commanded gain of the frame
      exptime_sec (float): exposure time of the frame
      frameid (int): an integer value used to indentify the frame

    Returns:
      filename (String): filename with path of the generated FITS file
    """
    signal = np.zeros(1)

    prhd, exthd = create_default_L1_headers()
    # Mock error maps
    err = np.ones(1)
    dq = np.zeros(1, dtype = np.uint16)
    # Creating a FITS file to assign it a filename with the frame ID
    prim = fits.PrimaryHDU(header = prhd)
    hdr_img = fits.ImageHDU(signal, header=exthd)
    hdul = fits.HDUList([prim, hdr_img])
    # Record actual commanded EM
    hdul[1].header['EMGAIN_C'] = cmdgain
    # Record actual exposure time
    hdul[1].header['EXPTIME'] = exptime_sec
    # Add corresponding VISTYPE
    hdul[0].header['VISTYPE'] = 'PUPILIMG'
    # IIT filename convention. TODO: replace with latest L1 filename version
    filename = str(Path('simdata', f'CGI_EXCAM_L1_{frameid:0{10}d}.fits'))
    hdul.writeto(filename, overwrite = True)
    return filename

def setup_module():
    global EXPTIME_MEAN_FRAME, NFRAMES_MEAN_FRAME
    global EXPTIME_KGAIN,  NFRAMES_KGAIN
    global EXPTIME_NONLIN, CMDGAIN_NONLIN
    global EXPTIME_EMGAIN_SEC, EXPTIME_EMGAIN_SEC
    global n_mean_frame_total, n_kgain_total, n_nonlin_wo_change_total, n_nonlin_w_change_total
    global dataset_w_change, dataset_wo_change 
    # Note: the values for the non-unity em gains and the number
    # of frames used for the mean frame, K-gain, non-linearity and EM-gain vs DAC
    # calibration come from either TVAC or some preliminary version of the
    # Commissioning test calculations
    
    # Global constants
    # Mean frame
    EXPTIME_MEAN_FRAME = 5
    NFRAMES_MEAN_FRAME = 30
    # Checks
    if NFRAMES_MEAN_FRAME < 30:
        raise Exception(f'Insufficient frames ({NFRAMES_MEAN_FRAME}) for the mean frame')
    
    # K-gain
    EXPTIME_KGAIN = [0.077, 0.770, 1.538, 2.308, 3.077, 3.846, 4.615, 5.385, 6.154,
        6.923, 7.692, 8.462, 9.231, 10.000, 10.769, 11.538, 12.308, 13.077, 13.846,
        14.615, 15.385, 1.538]
    NFRAMES_KGAIN = 5
    # Checks
    if NFRAMES_KGAIN < 5:
        raise Exception(f'Insufficient frames ({NFRAMES_KGAIN}) per unique exposure time in k-gain')
    if len(EXPTIME_KGAIN) < 22:
        raise Exception(f'Insufficient unique exposure times ({len(EXPTIME_KGAIN)}) in k-gain')
    if np.all(np.sign(np.diff(EXPTIME_KGAIN[:-1])) == 1) is False:
        raise Exception('Exposure times in K-gain must be monotonic but for the last value')
    
    # Non-linearity
    EXPTIME_NONLIN = [0.076, 0.758, 1.515, 2.273, 3.031, 3.789, 4.546, 5.304, 6.062,
        6.820, 7.577, 8.335, 9.093, 9.851, 10.608, 11.366, 12.124, 12.881, 13.639,
        14.397, 15.155, 1.515]
    CMDGAIN_NONLIN = [1.65, 5.24, 8.60, 16.70, 27.50, 45.26, 87.50, 144.10, 237.26,
        458.70, 584.40]
    # Checks
    if len(EXPTIME_NONLIN) < 22:
        raise Exception(f'Insufficient frames ({len(EXPTIME_NONLIN)}) per unique EM value in non-linearity')
    if len(CMDGAIN_NONLIN) < 11:
        raise Exception(f'Insufficient values of distinct EM Values ({len(EXPTIME_NONLIN)}) in non-linearity')
    if np.sum(np.array(EXPTIME_NONLIN) == EXPTIME_NONLIN[-1]) != 2:
        raise Exception('Last exposure time must be repeated once')
    if len(set(EXPTIME_NONLIN)) != len(EXPTIME_NONLIN) - 1:
        raise Exception('Only one exposure time can be repeated in non-linearity')
    if EXPTIME_NONLIN[-1] in EXPTIME_NONLIN[0:5] is False:
        raise Exception('The last exposure time must be present at the beginning of the exposure times in non-linearity')
    if np.all(np.sign(np.diff(EXPTIME_NONLIN[:-1])) == 1) is False:
        raise Exception('Exposure times in Non-linearity must be monotonic but for the last value')
    
    # Notice the pairing between unity and non-unity gain frames
    EM_EMGAIN=[1.000, 1.000, 1.007, 1.015, 1.024, 1.035, 1.047, 1.060, 1.076, 1.094,
        1.115, 1.138, 1.165, 1.197, 1.234, 1.276, 1.325, 1.385, 1.453, 1.534, 1.633,
        1.749, 1.890, 2.066, 2.278, 2.541, 2.873, 3.308, 3.858, 4.581, 5.577, 6.189,
        6.906, 7.753, 8.757, 9.955, 11.392, 13.222, 15.351, 17.953, 21.157, 25.128,
        30.082, 36.305, 44.621, 54.768, 67.779, 84.572, 106.378, 134.858, 172.244,
        224.385, 290.538, 378.283, 494.762, 649.232, 853.428]
    EXPTIME_EMGAIN_SEC=[5, 10, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5,
        5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 10, 10, 10,
        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10]
    NFRAMES_EMGAIN=[3, 5, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
        3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 5, 5, 5,
        5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5]
    # Checks
    if len(EM_EMGAIN) < 56:
        raise Exception(f'Insufficient number of EM gain values ({len(EM_EMGAIN)}) in EM-gain vs DAC')
    if len(EXPTIME_EMGAIN_SEC) != len(EM_EMGAIN):
        raise Exception(f'Inconsistent number of sets in EM-gain vs DAC')
    if len(EXPTIME_EMGAIN_SEC) != len(NFRAMES_EMGAIN):
        raise Exception(f'Inconsistent number of sets in EM-gain vs DAC')
    
    # Test data: Consider two possible scenarios: identical exposure times among the
    # different subsets of non-unity gain used to calibrate non-linearity or different
    # values of exposure times
    
    # Values for mean frame
    cmdgain_mean_frame, exptime_mean_frame = get_cmdgain_exptime_mean_frame(
        exptime_sec=EXPTIME_MEAN_FRAME,
        nframes=NFRAMES_MEAN_FRAME,
        )
    
    if len(cmdgain_mean_frame) != len(exptime_mean_frame):
        raise Exception('Inconsistent lengths in the mean frame')
    # Total number of frames
    n_mean_frame_total = len(cmdgain_mean_frame)
    
    # Values for K-gain
    cmdgain_kgain, exptime_kgain = get_cmdgain_exptime_kgain(
        exptime_sec=EXPTIME_KGAIN,
        nframes=NFRAMES_KGAIN,
        )
    if len(cmdgain_kgain) != len(exptime_kgain):
        raise Exception('Inconsistent lengths in k-gain')
    # Total number of frames
    n_kgain_total = len(cmdgain_kgain)
    
    # Values for Non-linearity
    cmdgain_nonlin_wo_change, exptime_nonlin_wo_change = get_cmdgain_exptime_nonlin(
        exptime_sec=EXPTIME_NONLIN,
        nonunity_em=CMDGAIN_NONLIN,
        change_exptime=False,
        )
    if len(cmdgain_nonlin_wo_change) != len(exptime_nonlin_wo_change):
        raise Exception('Inconsistent lengths in non-linearity')
    # Total number of frames
    n_nonlin_wo_change_total = len(cmdgain_nonlin_wo_change)
    
    cmdgain_nonlin_w_change, exptime_nonlin_w_change = get_cmdgain_exptime_nonlin(
        exptime_sec=EXPTIME_NONLIN,
        nonunity_em=CMDGAIN_NONLIN,
        change_exptime=True,
        )
    if len(cmdgain_nonlin_w_change) != len(exptime_nonlin_w_change):
        raise Exception('Inconsistent lengths in non-linearity')
    # Total number of frames
    n_nonlin_w_change_total = len(cmdgain_nonlin_w_change)
    
    # Values for EM-gain vs DAC
    cmdgain_emgain, exptime_emgain = get_cmdgain_exptime_emgain(
        em_emgain = EM_EMGAIN,
        exptime_emgain_sec = EXPTIME_EMGAIN_SEC,
        nframes = NFRAMES_EMGAIN,
        )
    if len(cmdgain_emgain) != len(exptime_emgain):
        raise Exception(f'Inconsistent lengths in em-gain vs dac')
    # Total number of frames
    n_emgain_total = len(cmdgain_emgain)
    
    # DRP Dataset
    # Create directory for temporary data files (not tracked by git)
    if not os.path.exists(Path('simdata')):
        os.mkdir(Path('simdata'))
    
    idx_frame = 0
    filename_list = []
    # Mean frame
    print('Generating frames for mean frame')
    for i_f in range(n_mean_frame_total):
        filename = make_minimal_image(
            cmdgain=cmdgain_mean_frame[i_f],
            exptime_sec=exptime_mean_frame[i_f],
            frameid=idx_frame,
            )
        filename_list += [filename]
        idx_frame += 1
    # K-gain
    print('Generating frames for k-gain')
    for i_f in range(n_kgain_total):
        filename = make_minimal_image(
            cmdgain=cmdgain_kgain[i_f],
            exptime_sec=exptime_kgain[i_f],
            frameid=idx_frame,
            )
        filename_list += [filename]
        idx_frame += 1
    # EM-gain
    print('Generating frames for em-gain')
    for i_f in range( n_emgain_total):
        filename = make_minimal_image(
            cmdgain=cmdgain_emgain[i_f],
            exptime_sec=exptime_emgain[i_f],
            frameid=idx_frame,
            )
        filename_list += [filename]
        idx_frame += 1
    # Non-linearity (two cases)
    print('Generating frames for non-linearity')
    filename_wo_change_list = copy.deepcopy(filename_list)
    for i_f in range(n_nonlin_wo_change_total):
        filename = make_minimal_image(
            cmdgain=cmdgain_nonlin_wo_change[i_f],
            exptime_sec=exptime_nonlin_wo_change[i_f],
            frameid=idx_frame,
            )
        filename_wo_change_list += [filename]
        idx_frame += 1
    
    filename_w_change_list = copy.deepcopy(filename_list)
    for i_f in range(n_nonlin_w_change_total):
        filename = make_minimal_image(
            cmdgain=cmdgain_nonlin_wo_change[i_f],
            exptime_sec=exptime_nonlin_wo_change[i_f],
            frameid=idx_frame,
            )
        filename_w_change_list += [filename]
        idx_frame += 1
    
    # Shuffle file order randomnly
    random.shuffle(filename_wo_change_list)
    random.shuffle(filename_w_change_list)
    
    # Create datasets
    dataset_wo_change = data.Dataset(filename_wo_change_list)
    dataset_w_change = data.Dataset(filename_w_change_list)
    
    # Delete temporary test FITS 
    for filepath in filename_wo_change_list:
        os.remove(filepath)
    for filepath in filename_w_change_list:
        # Delete remaining non-linearity FITS
        try:
            os.remove(filepath)
        except:
            pass

def test_kgain_sorting():
    """
    Apply the sorting algorithm to a dataset for K-gain and non-linearity
    calibration including EM-gain calibration files in the set to obtain
    the dataset needed for K-gain calibration and check the resulting
    dataset is consistent with the input dataset. K-gain uses unity gain
    frames only. No need to test both non-linearity subsets of data.
    """
    dataset_kgain = sorting.sort_pupilimg_frames(dataset_wo_change, cal_type='k-gain')

    # Checks
    n_mean_frame = 0
    n_kgain_test = 0
    filename_kgain_list = []
    exptime_mean_frame_list = []
    exptime_kgain_list = []
    # This way there's no need to perform a sum check and identifies any issue
    for idx_frame, frame in enumerate(dataset_kgain):
        if frame.pri_hdr['OBSNAME'] == 'MNFRAME':
            n_mean_frame += 1
            exptime_mean_frame_list += [frame.ext_hdr['EXPTIME']]
        elif frame.pri_hdr['OBSNAME'] == 'KGAIN':
            n_kgain_test += 1
            filename_kgain_list += [frame.filename]
            exptime_kgain_list += [frame.ext_hdr['EXPTIME']]
        else:
            try:
                raise Exception((f'Frame #{idx_frame}: Misidentified calibration' +
                   f"type in the calibration dataset. OBSNAME={frame.pri_hdr['OBSNAME']}"))
            except:
                raise Exception((f'Frame #{idx_frame}: Unidentified calibration',
                    'type in the Kgain calibration dataset'))
    
    # Same number of files as expected
    assert n_kgain_test == n_kgain_total
    # Unique exposure time for the mean frame
    assert len(set(exptime_mean_frame_list)) == 1
    # Expected exposure time for the mean frame
    assert exptime_mean_frame_list[0] == EXPTIME_MEAN_FRAME
    # Expected number of frames for the mean frame
    assert n_mean_frame == NFRAMES_MEAN_FRAME
    # Expected identical number of frames per exposure time in K-gain with
    # only one repeated case at the end
    kgain_unique, kgain_counts = np.unique(exptime_kgain_list, return_counts=True)
    assert len(set(kgain_counts)) == 2
    assert min(kgain_counts) == NFRAMES_KGAIN
    assert max(kgain_counts) == 2*min(kgain_counts)
    # Needs ordering
    idx_kgain_sort = np.argsort(filename_kgain_list)
    # Expected exposure times for K-gain
    exptime_kgain_arr = np.array(exptime_kgain_list)[idx_kgain_sort]
    assert len(set(exptime_kgain_arr[-NFRAMES_KGAIN:])) == 1
    assert exptime_kgain_arr[-1] in exptime_kgain_arr[0:-NFRAMES_KGAIN]

def test_nonlin_sorting_wo_change():
    """
    Apply the sorting algorithm to a dataset for K-gain and non-linearity
    calibration including EM-gain calibration files in the set to obtain
    the dataset needed for non-linearity calibration and check the
    resulting dataset is consistent with the input dataset. This test has
    identical exposure times among the different subsets of non-unity gain
    used to calibrate non-linearity
    """        
    dataset_nonlin_wo_change = sorting.sort_pupilimg_frames(dataset_wo_change, cal_type='non-lin')

    # Checks
    n_mean_frame = 0
    n_nonlin_test = 0
    filename_nonlin_list = []
    exptime_mean_frame_list = []
    exptime_nonlin_list = []
    cmdgain_nonlin_list = []
    # This way there's no need to perform a sum check and identifies any issue
    for idx_frame, frame in enumerate(dataset_nonlin_wo_change):
        if frame.pri_hdr['OBSNAME'] == 'MNFRAME':
            n_mean_frame += 1
            exptime_mean_frame_list += [frame.ext_hdr['EXPTIME']]
        elif frame.pri_hdr['OBSNAME'] == 'NONLIN':
            n_nonlin_test += 1
            filename_nonlin_list += [frame.filename]
            exptime_nonlin_list += [frame.ext_hdr['EXPTIME']]
            cmdgain_nonlin_list += [frame.ext_hdr['EMGAIN_C']]
        # Testing only non-unity gain frames for Non-linearity
        elif frame.pri_hdr['OBSNAME'] == 'KGAIN':
            pass
        else:
            try:
                raise Exception((f'Frame #{idx_frame}: Misidentified calibration' +
                   f"type in the calibration dataset. OBSNAME={frame.pri_hdr['OBSNAME']}"))
            except:
                raise Exception((f'Frame #{idx_frame}: Unidentified calibration',
                    'type in the Non-linearity calibration dataset'))
    
    # Same number of files as expected
    assert n_nonlin_test == n_nonlin_wo_change_total
    # Unique exposure time for the mean frame
    assert len(set(exptime_mean_frame_list)) == 1
    # Expected exposure time for the mean frame
    assert exptime_mean_frame_list[0] == EXPTIME_MEAN_FRAME
    # Expected number of frames for the mean frame
    assert n_mean_frame == NFRAMES_MEAN_FRAME
    # Needs ordering
    idx_nonlin_sort = np.argsort(filename_nonlin_list)
    # Expected exposure times for Non-linearity
    exptime_nonlin_arr = np.array(exptime_nonlin_list)[idx_nonlin_sort]
    # Subset of unique non-unity EM gains
    nonlin_em_gain_arr = np.unique(cmdgain_nonlin_list)
    nonlin_em_gain_arr.sort()
    assert np.all(nonlin_em_gain_arr == CMDGAIN_NONLIN)
    n_exptime_nonlin = len(EXPTIME_NONLIN)
    # Expected exposure times (this test keeps the same values for all subsets) 
    exptime_nonlin_arr = np.array(exptime_nonlin_list)[idx_nonlin_sort]
    for idx_em, nonlin_em in enumerate(nonlin_em_gain_arr):
        assert np.all(exptime_nonlin_arr[idx_em*n_exptime_nonlin:(idx_em+1)*n_exptime_nonlin] == EXPTIME_NONLIN)
        assert (exptime_nonlin_arr[(idx_em+1)*n_exptime_nonlin-1] in
            exptime_nonlin_arr[idx_em*n_exptime_nonlin:(idx_em+1)*n_exptime_nonlin-1])

def test_nonlin_sorting_w_change():
    """
    Apply the sorting algorithm to a dataset for K-gain and non-linearity
    calibration including EM-gain calibration files in the set to obtain
    the dataset needed for non-linearity calibration and check the
    resulting dataset is consistent with the input dataset. This test has
    different exposure times among the different subsets of non-unity gain
    used to calibrate non-linearity
    """
    dataset_nonlin_w_change = sorting.sort_pupilimg_frames(dataset_w_change, cal_type='non-lin')

    # Checks
    n_mean_frame = 0
    n_nonlin_test = 0
    filename_nonlin_list = []
    exptime_mean_frame_list = []
    exptime_nonlin_list = []
    cmdgain_nonlin_list = []
    # This way there's no need to perform a sum check and identifies any issue
    for idx_frame, frame in enumerate(dataset_nonlin_w_change):
        if frame.pri_hdr['OBSNAME'] == 'MNFRAME':
            n_mean_frame += 1
            exptime_mean_frame_list += [frame.ext_hdr['EXPTIME']]
        elif frame.pri_hdr['OBSNAME'] == 'NONLIN':
            n_nonlin_test += 1
            filename_nonlin_list += [frame.filename]
            exptime_nonlin_list += [frame.ext_hdr['EXPTIME']]
            cmdgain_nonlin_list += [frame.ext_hdr['EMGAIN_C']]
        # Testing only non-unity gain frames for Non-linearity
        elif frame.pri_hdr['OBSNAME'] == 'KGAIN':
            pass
        else:
            try:
                raise Exception((f'Frame #{idx_frame}: Misidentified calibration' +
                   f"type in the calibration dataset. OBSNAME={frame.pri_hdr['OBSNAME']}"))
            except:
                raise Exception((f'Frame #{idx_frame}: Unidentified calibration',
                    'type in the Non-linearity calibration dataset'))

    # Same number of files as expected
    assert n_nonlin_test == n_nonlin_wo_change_total
    # Unique exposure time for the mean frame
    assert len(set(exptime_mean_frame_list)) == 1
    # Expected exposure time for the mean frame
    assert exptime_mean_frame_list[0] == EXPTIME_MEAN_FRAME
    # Expected number of frames for the mean frame
    assert n_mean_frame == NFRAMES_MEAN_FRAME
    # Needs ordering
    idx_nonlin_sort = np.argsort(filename_nonlin_list)
    # Expected exposure times for Non-linearity
    exptime_nonlin_arr = np.array(exptime_nonlin_list)[idx_nonlin_sort]
    # Subset of unique non-unity EM gains
    nonlin_em_gain_arr = np.unique(cmdgain_nonlin_list)
    nonlin_em_gain_arr.sort()
    assert np.all(nonlin_em_gain_arr == CMDGAIN_NONLIN)
    n_exptime_nonlin = len(EXPTIME_NONLIN)
    # Expected exposure times (this test changed the values for all subsets)
    exptime_nonlin_arr = np.array(exptime_nonlin_list)[idx_nonlin_sort]
    for idx_em, nonlin_em in enumerate(nonlin_em_gain_arr):
        assert (exptime_nonlin_arr[(idx_em+1)*n_exptime_nonlin-1] in
            exptime_nonlin_arr[idx_em*n_exptime_nonlin:(idx_em+1)*n_exptime_nonlin-1])
    
if __name__ == "__main__":
    print('Running test_sort_pupilimg_sorting')
    # Testing the sorting algorithm for K-gain calibration
    test_kgain_sorting()
    print('* K-gain tests passed')

    # Testing the sorting algorithm for non-linearity calibration
    test_nonlin_sorting_wo_change()
    print('* Non-linearity tests with identical exposure times among non-unity gains passed')

    test_nonlin_sorting_w_change()
    print('* Non-linearity tests with different exposure times among non-unity gains passed')




"""Photon count a stack of analog images and return a mean expected electron 
count array with photometric corrections.

"""
import warnings
import numpy as np
from astropy.io import fits
import corgidrp.data as data
from corgidrp import check

def varL23(g, L, T, N):
    '''Expected variance after photometric correction.  
    See https://doi.org/10.1117/1.JATIS.9.4.048006 for details.
    
    Args:
        g (scalar): EM gain
        L (2-D array): mean expected number of electrons
        T (scalar): threshold 
        N (2-D array): number of frames

    Returns:
        (float): variance from photon counting and the photometric correction

    '''
    Const = 6/(6 + L*(6 + L*(3 + L)))
    eThresh = (Const*(np.e**(-T/g)*L*(2*g**2*(6 + L*(3 + L)) +
            2*g*L*(3 + L)*T + L**2*T**2))/(12*g**2))
    std_dev = np.sqrt(N * eThresh * (1-eThresh))

    return (std_dev)**2*(((np.e**((T/g)))/N) +
    2*((np.e**((2*T)/g)*(g - T))/(2*g*N**2))*(N*eThresh) +
    3*(((np.e**(((3*T)/g)))*(4*g**2 - 8*g*T + 5*T**2))/(
    12*g**2*N**3))*(N*eThresh)**2)**2

class PhotonCountException(Exception):
    """Exception class for photon_count module."""


def photon_count(e_image, thresh):
    """Convert an analog image into a photon-counted image.

    Args: 
        e_image (array_like, float): Analog image (e-).
        thresh (float): Photon counting threshold (e-). Values > thresh will be assigned a 1,
            values <= thresh will be assigned a 0.

    Returns:
        pc_image (array_like, float): Output digital image in units of photons.

    B Nemati and S Miller - UAH - 06-Aug-2018

    """
    # Check if input is an array/castable to one
    e_image = np.array(e_image).astype(float)
    if len(e_image.shape) == 0:
        raise PhotonCountException('e_image must have length > 0')

    pc_image = np.zeros(e_image.shape, dtype=int)
    pc_image[e_image > thresh] = 1

    return pc_image


def get_pc_mean(input_dataset, pc_master_dark=None, T_factor=None, pc_ecount_max=None, 
                niter=2, mask_filepath=None, safemode=True, inputmode='illuminated', bin_size=None):
    """Take a stack of images, frames of the same exposure 
    time, k gain, read noise, and EM gain, and return the mean expected value per 
    pixel. The frames are taken in photon-counting mode, which means high EM 
    gain and very low exposure time.  

    The frames in each stack should be SCI image (1024x1024) frames that have 
    had some of the L2b steps applied:

    - have had their bias subtracted
    - have had masks made for cosmic rays
    - have been corrected for nonlinearity (These first 3 steps make the frames L2a.)
    - have been frame-selected (to weed out bad frames)
    - have been converted from DN to e-

    This algorithm will photon count each frame individually,
    then co-add the photon-counted frames. The co-added frame is then averaged
    and corrected for thresholding and coincidence loss, returning the mean
    expected array in units of photoelectrons if dark-subtracted (detected electrons 
    if not dark-subtracted).  The threshold is determined by the input "T_factor", 
    and the value stored in DetectorParams is used if this input is None. 
    
    This function can be used for photon-counting illuminated and dark datasets 
    (see Args below). 

    Args:
        input_dataset (corgidrp.data.Dataset): This is an instance of corgidrp.data.Dataset containing the 
            frames to photon-count. All the frames must have the same 
            exposure time, EM gain, k gain, and read noise.  If the input dataset's header key 'VISTYPE' is equal to 'DARK', a 
            photon-counted master dark calibration product will be produced.  Otherwise, the input dataset is assumed to consist of illuminated 
            frames intended for photon-counting.
        pc_master_dark (corgidrp.data.Dark): Dark containing photon-counted master dark(s) to be used for dark subtraction.  There is a 3-D cube 
            of master darks, 1 2-D slice per subset of frames specified by the input bin_size (see below).
            If None, no dark subtraction is done.
        T_factor (float): The number of read noise standard deviations at which to set the threshold for photon counting.  If None, the value is drawn from corgidrp.data.DetectorParams. 
            Defaults to None.
        pc_ecount_max (float): Maximum allowed electrons/pixel/frame for photon counting.  If None, the value is drawn from corgidrp.data.DetectorParams. 
            Defaults to None.
        niter (int, optional): Number of Newton's method iterations (used for photometric 
            correction). Defaults to 2.
        mask_filepath (str): Filepath to a .fits file with a pixel mask in the default HDU.  The mask should be an array of the same shape as that found in each frame of the input_dataset, and the mask 
            should be 0 where the region of interest is and 1 for pixels to be masked (not considered).
        safemode (bool): If False, the function does not halt due to an exception (useful for the iterative process of digging the dark hole).  
            If True, the function halts with an exception if the mean intensity of the unmasked pixels (or all pixels if no mask provided) is greater than pc_ecount_max or if the minimum photon-counted pixel value is negative.
            If False, the function gives a warning for these instead.
            Defaults to True.
        inputmode (str):  If 'illuminated', the frames are assumed to be illuminated frames.  If 'darks', frames are assumed to be dark frames input for creation of a photon-counted master dark. 
            This flag shows the user's intention with the input, and this input is checked against the file type of the dataset for compatibility (e.g., if this input is 'darks' while 'VISTYPE' is not equal 
            to 'DARK', an exception is raised).
        bin_size (int):  If one wishes to break up the input dataset into subsets of frames to photon-count separately (e.g., for testing the balancing act between 
            good SNR with many frames vs countering speckle time variability with fewer frames), one specifies this number for the size of each subset. If the number does not evenly divide the 
            number of frames in input_dataset, the remainder frames are ignored.  The output is the a dataset containing the more than 1 photon-counted mean-combined frame. Defaults to None, in which case 
            the entire input dataset is used, and the output dataset consists of one frame.

    Returns:
        corgidrp.data.Dataset or corgidrp.data.Dark: If If the input dataset's header key 'VISTYPE' is not equal to 'DARK', 
            corgidrp.data.Dataset is the output type, and the output is the processed illuminated set, whether 
            dark subtraction happened or not.  Contains mean expected array (detected electrons if not dark-subtracted, 
            photoelectrons if dark-subtracted). 
            If the input dataset's header key 'VISTYPE' is equal to 'DARK', corgidrp.data.Dark is the output type, and the output
            is the processed dark set.  Contains mean expected array (detected electrons).

    References
    ----------
    [1] https://www.spiedigitallibrary.org/conference-proceedings-of-spie/11443/114435F/Photon-counting-and-precision-photometry-for-the-Roman-Space-Telescope/10.1117/12.2575983.full
    [2] https://doi.org/10.1117/1.JATIS.9.4.048006

    B Nemati, S Miller - UAH - 13-Dec-2020
    Kevin Ludwick - UAH - 2023

    """
    if not isinstance(niter, (int, np.integer)) or niter < 1:
            raise PhotonCountException('niter must be an integer greater than '
                                        '0')
    if bin_size is None:
        bin_size = len(input_dataset)
    check.positive_scalar_integer(bin_size, 'bin_size', TypeError)
    if bin_size > len(input_dataset):
        raise ValueError('bin_size must be less than the number of frames in input_dataset.')
        
    num_bins = len(input_dataset)//bin_size 
    
    list_new_image = []
    list_err = [] # only used for dark processing case
    list_dq = [] # only used for dark processing case
    index_of_last_frame_used = num_bins*(len(input_dataset)//num_bins)
    # this for loop ignores the remainder from the division above
    for i in range(num_bins):
        subset_frames = input_dataset.frames[bin_size*i:bin_size*(i+1)]
        sub_dataset = data.Dataset(subset_frames)

        test_dataset, _ = sub_dataset.copy().split_dataset(exthdr_keywords=['EXPTIME', 'EMGAIN_C', 'KGAINPAR', 'RN'])
        if len(test_dataset) > 1:
            raise PhotonCountException('All frames must have the same exposure time, '
                                    'commanded EM gain, and k gain.')
        datasets, val = test_dataset[0].split_dataset(prihdr_keywords=['VISTYPE'])
        if len(val) != 1:
            raise PhotonCountException('There should only be 1 \'VISTYPE\' value for the dataset.')
        if val[0] == 'DARK':
            if inputmode != 'darks':
                raise PhotonCountException('Inputmode is not \'darks\', but the input dataset has \'VISTYPE\' = \'DARK\'.')
            if pc_master_dark is not None:
                raise PhotonCountException('The input frames are \'VISTYPE\'=\'DARK\' frames, so no pc_master_dark should be provided.')
        if val[0] != 'DARK':
            if inputmode != 'illuminated':
                raise PhotonCountException('Inputmode is not \'illuminated\', but the input dataset has \'VISTYPE\' not equal to \'DARK\'.')
        if 'ISPC' in datasets[0].frames[0].ext_hdr:
            if datasets[0].frames[0].ext_hdr['ISPC'] != True:
                raise PhotonCountException('\'ISPC\' header value must be True if these frames are to be processed as photon-counted.')

        dataset = datasets[0]
        
        pc_means = []
        errs = []
        dqs = []
        # getting number of read noise standard deviations at which to set the
        # photon-counting threshold
        if T_factor is None:
            detector_params = data.DetectorParams({})
            T_factor = detector_params.params['TFACTOR']
        # getting maximum allowed electrons/pixel/frame for photon counting
        if pc_ecount_max is None:
            detector_params = data.DetectorParams({})
            pc_ecount_max = detector_params.params['PCECNTMX']
        if mask_filepath is None:
            mask = np.zeros_like(dataset.frames[0].data)
        else:
            mask = fits.getdata(mask_filepath)
        
        considered_indices = (mask==0)

        # now get threshold to use for photon-counting
        read_noise = test_dataset[0].frames[0].ext_hdr['RN']
        thresh = T_factor*read_noise
        if thresh < 0:
            raise PhotonCountException('thresh must be nonnegative')
        
        if 'EMGAIN_M' in dataset.frames[0].ext_hdr: # if EM gain measured directly from frame TODO change hdr name if necessary
            em_gain = dataset.frames[0].ext_hdr['EMGAIN_M']
        else:
            em_gain = dataset.frames[0].ext_hdr['EMGAIN_A']
            if em_gain > 0: # use applied EM gain if available
                em_gain = dataset.frames[0].ext_hdr['EMGAIN_A']
            else: # use commanded gain otherwise
                em_gain = dataset.frames[0].ext_hdr['EMGAIN_C']

        if thresh >= em_gain:
            if safemode:
                raise Exception('thresh should be less than em_gain for effective '
                'photon counting')
            else:
                warnings.warn('thresh should be less than em_gain for effective '
                'photon counting')
        if np.nanmean(dataset.all_data[:, considered_indices]/em_gain) > pc_ecount_max:
            if safemode:
                raise Exception('average # of electrons/pixel is > pc_ecount_max, which means '
                'the average # of photons/pixel may be > pc_ecount_max (depending on the QE).  Can decrease frame '
                'time to get lower average # of photons/pixel.')
            else:
                warnings.warn('average # of electrons/pixel is > pc_ecount_max, which means '
                'the average # of photons/pixel may be > pc_ecount_max (depending on the QE).  Can decrease frame '
                'time to get lower average # of photons/pixel.')
        if read_noise <=0:
            raise Exception('read noise should be greater than 0 for effective '
            'photon counting')
        if thresh < 4*read_noise: # leave as just a warning
            warnings.warn('thresh should be at least 4 or 5 times read_noise for '
            'accurate photon counting')

        # Photon count stack of frames
        frames_pc = photon_count(dataset.all_data, thresh)
        bool_map = dataset.all_dq.astype(bool).astype(float)
        bool_map[bool_map > 0] = np.nan
        bool_map[bool_map == 0] = 1
        nframes = np.nansum(bool_map, axis=0)
        # upper and lower bounds for PC (for getting accurate err)
        frames_pc_up = photon_count(dataset.all_data+dataset.all_err[:,0], thresh)
        frames_pc_low = photon_count(dataset.all_data-dataset.all_err[:,0], thresh)
        frames_pc_masked = frames_pc * bool_map
        frames_pc_masked_up = frames_pc_up * bool_map
        frames_pc_masked_low = frames_pc_low * bool_map
        # Co-add frames
        frame_pc_coadded = np.nansum(frames_pc_masked, axis=0)
        frame_pc_coadded_up = np.nansum(frames_pc_masked_up, axis=0)
        frame_pc_coadded_low = np.nansum(frames_pc_masked_low, axis=0)
        
        # Correct for thresholding and coincidence loss; any pixel masked all the 
        # way through the stack may give NaN, but it should be masked in lam_newton_fit(); 
        # and it doesn't matter anyways since its DQ value will be 1 (it will be masked when the 
        # bad pixel correction is run, which comes after this photon-counting step)
        mean_expected = corr_photon_count(frame_pc_coadded, nframes, thresh,
                                            em_gain, considered_indices, niter)
        mean_expected_up = corr_photon_count(frame_pc_coadded_up, nframes, thresh,
                                            em_gain, considered_indices, niter)
        mean_expected_low = corr_photon_count(frame_pc_coadded_low, nframes, thresh,
                                            em_gain, considered_indices, niter)
        ##### error calculation: accounts for err coming from input dataset and 
        # statistical error from the photon-counting and photometric correction process. 
        # expected error from photon counting (biggest source from the actual values, not 
        # mean_expected_up or mean_expected_low):
        pc_variance = varL23(em_gain,mean_expected,thresh,nframes)
        up = mean_expected_up +  pc_variance
        low = mean_expected_low -  pc_variance
        errs.append(np.max([up - mean_expected, mean_expected - low], axis=0))
        dq = (nframes == 0).astype(int) 
        pc_means.append(mean_expected)
        dqs.append(dq)
        
        if pc_master_dark is not None:
            if type(pc_master_dark) is not data.Dark:
                raise Exception('Input type for pc_master_dark must be a Dataset of corgidrp.data.Dark instances.')
            if 'PC_STAT' not in pc_master_dark.ext_hdr:
                raise PhotonCountException('\'PC_STAT\' must be a key in the extension header of each frame of pc_master_dark.')
            if pc_master_dark.ext_hdr['PC_STAT'] != 'photon-counted master dark':
                raise PhotonCountException('Each frame of pc_master_dark must be a photon-counted master dark (i.e., '
                                        'the extension header key \'PC_STAT\' must be \'photon-counted master dark\').')
            if 'PCTHRESH' not in pc_master_dark.ext_hdr:
                raise PhotonCountException('Threshold should be stored under the header \'PCTHRESH\'.')
            if pc_master_dark.ext_hdr['PCTHRESH'] != thresh:
                raise PhotonCountException('Threshold used for photon-counted master dark should match the threshold to be used for the illuminated frames.')
            if pc_master_dark.ext_hdr['NUM_FR'] < len(sub_dataset):
                raise PhotonCountException('Number of frames that created the photon-counted master dark must be greater than or equal to the number of illuminated frames in order for the result to be reliable.')
            # in case the number of subsets of darks < number of subsets of brights, which can happen since the number of darks within a subset can be bigger than the number in a bright subset
            j = np.mod(i, pc_master_dark.data.shape[0])
            pc_means.append(pc_master_dark.data[j])
            dqs.append(pc_master_dark.dq[j])
            errs.append(pc_master_dark.err[0][j])
            dark_sub = "yes"
        else:
            pc_means.append(np.zeros_like(pc_means[0]))
            dqs.append(np.zeros_like(pc_means[0]).astype(int))
            errs.append(np.zeros_like(pc_means[0]))
            dark_sub = "no"

        # now subtract the dark PC mean
        combined_pc_mean = pc_means[0] - pc_means[1]
        combined_pc_mean[combined_pc_mean<0] = 0
        combined_err = np.sqrt(errs[0]**2 + errs[1]**2)
        combined_dq = np.bitwise_or(dqs[0], dqs[1])
        pri_hdr = dataset[-1].pri_hdr.copy()
        ext_hdr = dataset[-1].ext_hdr.copy()
        err_hdr = dataset[-1].err_hdr.copy()
        dq_hdr = dataset[-1].dq_hdr.copy()
        hdulist = dataset[-1].hdu_list.copy()

        if val[0] != "DARK":  
            new_image = data.Image(combined_pc_mean, pri_hdr=pri_hdr, ext_hdr=ext_hdr, err=combined_err, dq=combined_dq, err_hdr=err_hdr, 
                                dq_hdr=dq_hdr, input_hdulist=hdulist) 
            new_image.filename = dataset[-1].filename.replace("L2a", "L2b")
            new_image.ext_hdr['PCTHRESH'] = thresh
            new_image.ext_hdr['NUM_FR'] = len(sub_dataset) 
            new_image._record_parent_filenames(sub_dataset) 
            list_new_image.append(new_image)
        else:
            list_new_image.append(combined_pc_mean)
            list_err.append(combined_err)
            list_dq.append(combined_dq)
    if val[0] != "DARK":
        pc_ill_dataset = data.Dataset(list_new_image)
        pc_ill_dataset.update_after_processing_step("Photon-counted {0} illuminated frames for each PC frame of the output dataset.  Number of subsets: {1}.  Total number of frames in input dataset: {2}. Using T_factor={3} and niter={4}. Dark-subtracted: {5}.".format(len(sub_dataset), num_bins, len(input_dataset), T_factor, niter, dark_sub))
        
        return pc_ill_dataset
    else:
        ext_hdr['PC_STAT'] = 'photon-counted master dark'
        ext_hdr['NAXIS1'] = combined_pc_mean.shape[0]
        ext_hdr['NAXIS2'] = combined_pc_mean.shape[1]
        ext_hdr['PCTHRESH'] = thresh
        ext_hdr['NUM_FR'] = len(sub_dataset) 
        ext_hdr['HISTORY'] = "Photon-counted {0} dark frames for each master dark of the output dataset.  Number of subsets: {1}.  Total number of master darks in input dataset: {2}. Using T_factor={3} and niter={4}.".format(len(sub_dataset), num_bins, len(input_dataset), T_factor, niter)
        pc_dark = data.Dark(np.stack(list_new_image), pri_hdr=pri_hdr, ext_hdr=ext_hdr, err=np.stack([list_err]), dq=np.stack(list_dq), err_hdr=err_hdr, input_dataset=input_dataset[:index_of_last_frame_used])
        return pc_dark

def corr_photon_count(nobs, nfr, t, g, mask_indices, niter=2):
    """Correct photon counted images.

    Args:
        nobs (array_like): Number of observations (Co-added photon-counted frame).
        nfr (int): Number of coadded frames, accounting for masked pixels in the frames.
        t (float): Photon-counting threshold.
        g (float): EM gain.
        mask_indices (array-like): indices of pixel positions to use.
        niter (int, optional): Number of Newton's method iterations. Defaults to 2.

    Returns:
        lam (array_like): Mean expeted electrons per pixel (lambda).

    """
    # Get an approximate value of lambda for the first guess of the Newton fit
    lam0 = calc_lam_approx(nobs, nfr, t, g)

    # Use Newton's method to converge at a value for lambda
    lam = lam_newton_fit(nobs, nfr, t, g, lam0, niter, mask_indices)

    return lam


def calc_lam_approx(nobs, nfr, t, g):
    """Approximate lambda calculation.

    This will calculate the first order approximation of lambda, and for values
    that are out of bounds (e.g. from statistical fluctuations) it will revert
    to the zeroth order.

    Args:
        nobs (array_like): Number of observations (Co-added photon counted frame).
        nfr (int): Number of coadded frames.
        t (float): Photon counting threshold.
        g (float): EM gain used when taking images.

    Returns:
        lam1 (array_like): Mean expected (lambda).

    """
    # First step of equation (before taking log)
    init = 1 - (nobs/nfr) * np.exp(t/g)
    # Mask out all values less than or equal to 0
    lam_m = np.zeros_like(init).astype(bool)
    lam_m[init > 0] = True

    # Use the first order approximation on all values greater than zero
    lam1 = np.zeros_like(init)
    lam1[lam_m] = -np.log(init[lam_m])

    # For any values less than zero, revert to the zeroth order approximation
    lam0 = nobs / nfr
    lam1[~lam_m] = lam0[~lam_m]

    return lam1


def lam_newton_fit(nobs, nfr, t, g, lam0, niter, mask_indices):
    """Newton fit for finding lambda.

    Args:
        nobs (array_like): Number of observations (Co-added photon counted frame).
        nfr (int): Number of coadded frames.
        t (float): Photon counting threshold.
        g (float): EM gain used when taking images.
        lam0 (array_like): Initial guess for lambda.
        niter (int): Number of Newton's fit iterations to take.
        mask_indices (array-like): indices of pixel positions to use.

    Returns:
        lam (array_like): Mean expected (lambda).

    """
    # Mask out zero values to avoid divide by zero
    lam_est_m = np.ma.masked_array(lam0, mask=(lam0==0))
    nobs_m = np.ma.masked_array(nobs, mask=(nobs==0))

    # Iterate Newton's method
    for i in range(niter):
        func = _calc_func(nobs_m, nfr, t, g, lam_est_m)
        dfunc = _calc_dfunc(nfr, t, g, lam_est_m)
        lam_est_m -= func / dfunc

    if np.nanmin(lam_est_m.data[mask_indices]) < 0:
        raise PhotonCountException('negative number of photon counts; '
        'try decreasing the frametime')

    # Fill zero values back in
    lam = lam_est_m.filled(0)

    return lam

def _calc_func(nobs, nfr, t, g, lam):
    """Objective function for lambda for Newton's method for all applying photometric correction.
    
    Args:
        nobs (array-like): number of frames per pixel that passed the threshold
        nfr (array-like): number of unmasked frames per pixel total
        t (float): threshold for photon counting
        g (float): EM gain
        lam (array-like): estimated mean expected electron count

    Returns:
        func (array-like): objective function

    """
    epsilon_prime = (lam*(2*g**2*(6 + lam*(3 + lam)) + 2*g*lam*(3 + lam)*t + 
            lam**2*t**2))/(2.*np.e**(t/g)*g**2*(6 + lam*(6 + lam*(3 + lam))))

    #if (nfr * epsilon_prime).any() > nobs.any():
    #    warnings.warn('Input photon flux is too high; decrease frametime')
    # This warning isn't necessary; could have a negative func but still
    # close enough to 0 for Newton's method
    func = nfr * epsilon_prime - nobs

    return func


def _calc_dfunc(nfr, t, g, lam):
    """Derivative with respect to lambda of objective function.
    
    Args:
        nfr (array-like): number of unmasked frames per pixel total
        t (float): threshold for photon counting
        g (float): EM gain
        lam (array-like): estimated mean expected electron count
    
    Returns:
        dfunc (array-like): derivative with respect to lambda of objective function
    """
    dfunc = (lam*nfr*(2*g**2*(3 + 2*lam) + 2*g*lam*t + 2*g*(3 + lam)*t + 
            2*lam*t**2))/(2.*np.e**(t/g)*g**2*(6 + lam*(6 + lam*(3 + lam)))) - (lam*(6 + 
            lam*(3 + lam) + lam*(3 + 2*lam))*nfr*
         (2*g**2*(6 + lam*(3 + lam)) + 2*g*lam*(3 + lam)*t + lam**2*t**2))/(2.*np.e**(t/g)*g**2*(6 + 
        lam*(6 + lam*(3 + lam)))**2) + (nfr*(2*g**2*(6 + lam*(3 + lam)) + 2*g*lam*(3 + lam)*t + 
        lam**2*t**2))/(2.*np.e**(t/g)*g**2*(6 + lam*(6 + lam*(3 + lam))))

    return dfunc